!Purpose:
!  Program for pairwise parentage assignments using marker data
!  Used as a subroutine called by the program AgeStructure for estimating Ne
!Input file:
!  AgeSexGtype.dat
!Record of revisions:
!  Date             Change
!  5/08/2009       Original code
!
!NOTE 1: EXE produced by FTN95 does not run well on 64bit cpus
!     2: For monoecious, diploid is assumed; For dioecious, diploid & haplodiploid assumed
!     3: For Linux/Mac, the binary must be POAssignment WITHOUT extension
!
!Compile command for optimal performance:
!  Salford FTN95        : ftn95 *.f90 /optimise /link
!  Intel ifort          : ifort *.f90 /O3 /Qipo /Qprec-div- /Qparallel /link /STACK:6400000
!  G95                  : g95 *.f90 -O3 -funroll-loops -fomit-frame-pointer
!
!Compile command for dianostics:
!g95 *.f90 -Wall -pedantic -fbounds-check -ftrace=full -std=f2003
!
!
MODULE VARIABLES
  IMPLICIT NONE
  INTEGER, PARAMETER :: I4B = SELECTED_INT_KIND(9)    ! -2147483648 ~ 2147483647
  INTEGER, PARAMETER :: I1B = SELECTED_INT_KIND(2)    ! -128 ~ 127
  INTEGER, PARAMETER :: SP = KIND(1.0)                ! Single Precision
  INTEGER, PARAMETER :: DP = KIND(1.0D0)              ! Double Precision
  REAL(DP), PARAMETER :: TinyValue = -1.E300_DP, Small = 1.E-300_DP
  INTEGER(I4B) :: NumLoci, SmplSize, MaxObsNumAllele
  INTEGER(I4B) :: NumKnownPrnt(2)
  INTEGER(I4B) :: MaxNumGenotype, FamilySizeDist, SamplingPrbKnown, NumCI95Sample
  INTEGER(I1B), ALLOCATABLE :: DominantMarker(:)
  INTEGER(I4B), ALLOCATABLE :: LocusError(:), GenotypeID2(:, :)
  Logical :: MarkerErrorFreeQ, AnyGtypeMissingQ
  INTEGER(I4B), ALLOCATABLE :: ObsNumAllele(:), NumAllele(:), &
    IndivParent(:, :)
  INTEGER(I4B), ALLOCATABLE :: IndivGType(:, :, :), CndGType(:, :, :, :)
  INTEGER(I4B), Target, ALLOCATABLE :: IndivGType2(:, :)
  INTEGER(I4B), Allocatable :: PopNumAllele(:), PopAllele(:, :)
  REAL(DP), Allocatable :: PopAlleleFre(:, :)
  REAL(DP), ALLOCATABLE :: ObsAlleleFre(:, :), &
    TransRate(:,:), DropRate(:), OtherErrorRate(:)
  INTEGER(I4B) :: NumCandidate(2), MeanNumCandidate(2)
  TYPE ExcludedList
    Integer(I4B) :: N
    Integer(I4B), Pointer :: List(:)
  END TYPE ExcludedList
  TYPE(ExcludedList), Allocatable :: ExcludedParent(:, :)
  REAL(DP) :: PropPrntSampled(2)
  REAL(DP), ALLOCATABLE :: CandidateLogL(:, :)
  CHARACTER(Len=500) :: FName, DataSetName
  TYPE FSProbPrt
    Integer(I4B), Pointer :: R(:)
  END TYPE FSProbPrt
  TYPE(FSProbPrt), Allocatable :: RawAllele(:)
  Character(Len = 20), Allocatable :: OffspringID(:), CndParentID(:, :), &
    MarkerID(:), SortCharacter(:)
  CHARACTER(Len = 999) :: ErrorMessage(5)
  Type PODyadProbType
    REAL(SP), Pointer :: P(:)
  END Type PODyadProbType
  TYPE(PODyadProbType), Allocatable :: PrbPOPair(:, :)
  REAL(SP), PARAMETER :: Tiny_SP = -1.E38_SP, Small_SP = 1.E-37_SP
  INTEGER(I4B), Allocatable :: IndivSex(:), &
    IndivAge(:), AgeSize(:, :)
  INTEGER(I4B) ::  NumAgeClass(0 : 2), MinParentAge(2)
  Type RealVector
    REAL(DP), Pointer :: R(:)
  END Type RealVector 
  TYPE(RealVector), Allocatable :: SamplingPrbPrior(:)      
END MODULE VARIABLES
!
Module SimuVariables
  USE VARIABLES
  IMPLICIT NONE
  REAL(DP), Allocatable :: CumulativeAlleleFre(:, :), MissingRate(:)
  REAL(DP) :: ThreShold(2, 10), ThreShold2(2, 10)
END Module SimuVariables
!
PROGRAM MAIN
  USE VARIABLES
  IMPLICIT NONE
  EXTERNAL :: DataProcess, InitialFSProb, Initialize, &
    PhenotypeGivenGenotype, ReadData, PairwiseInference  
  CALL ReadData
  CALL Initialize
  CALL DataProcess
  CALL PhenotypeGivenGenotype
  CALL InitialFSProb
  CALL PairwiseInference
END PROGRAM MAIN
!
SUBROUTINE PairwiseInference
!C Inferring pairwise sibship and parentage
  USE SimuVariables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L, N, IP(2)
  INTEGER(I4B), ALLOCATABLE :: Exclude(:) 
  REAL(DP) :: LogL(0 : 3), X, Y
  REAL(DP), ALLOCATABLE :: LogL1(:)
  CHARACTER(Len = 4) :: Confidence(0 : 2)
  External :: SingleParentLikelihood, &
    DetermineDeltaValue, SingleIndivLikelihood
  Allocate(LogL1(SmplSize))
!C Calculate individual offspring log likelihood
  DO I = 1, SmplSize
    CALL SingleIndivLikelihood(IndivGType(:, :, I), LogL1(I))
  END DO
  IF(ALL(NumCandidate(:) < 1)) RETURN
  CALL DetermineDeltaValue     ! determine threshold by simu
  CALL SingleParentLikelihood
  IF(Allocated(Exclude)) Deallocate(Exclude)
  I = Maxval(NumCandidate(1 : 2))    
  Allocate(Exclude(I))
  Exclude(:) = 0
  DO J = 1, 2
    IF(NumCandidate(J) < 1) CYCLE
    IF(J == 1)THEN
      OPEN(10, FILE = TRIM(FName) // '.PairwisePaternity')
    ELSE
      OPEN(10, FILE = TRIM(FName) // '.PairwiseMaternity')
    END IF
    WRITE(10, '(3A21)')'OffspringID', 'CandidateID', 'Confidence'
    N = J
    DO I = 1, SmplSize 
      IF(IndivAge(I) > NumAgeClass(J) - MinParentAge(J)) CYCLE
      IP(J) = 0
      IF(IndivParent(I, J) > 0) THEN
        IP(J) = IndivParent(I, J)
        L = 0
      ELSE
        Y = LogL1(I)
        LogL(0 : 1) = 0._DP
        IF(ExcludedParent(I, J)%N > 0) THEN
          Where(Exclude(: NumCandidate(J)) == 1) Exclude(: NumCandidate(J)) = 0
          DO K = 1, ExcludedParent(I, J)%N
            Exclude(ExcludedParent(I, J)%List(K)) = 1
          END DO
        END IF
        DO K = 1, NumCandidate(J)
          IF(IndivSex(K) /= J .OR. &
            IndivAge(K) - MinParentAge(J) < IndivAge(I)) CYCLE
          IF(ExcludedParent(I, J)%N > 0) THEN
            IF(Exclude(K) == 1) CYCLE
          END IF
          LogL(2) = PrbPOPair(I, N)%P(K) 
          X = LogL(2) - (Y + CandidateLogL(K, J))
          IF(X > LogL(0)) THEN
            LogL(1) = LogL(0)
            LogL(0) = X
            IP(J) = K
          ELSE IF(X > LogL(1))THEN
            LogL(1) = X
          END IF
        END DO
        IF(IndivParent(I, 3 - J) < 1)THEN       
          IF(LogL(0) - LogL(1) < ThreShold(J, 10)) THEN
            IP(J) = 0
          ELSE
            DO K = 9, 1, -1 
              IF(LogL(0) - LogL(1) < ThreShold(J, K)) THEN
                L = K + 1
                EXIT
              END IF
            END DO
            IF(LogL(0) - LogL(1) >= ThreShold(J, 1)) L = 1
          END IF
        ELSE
          IF(LogL(0) - LogL(1) < ThreShold2(J, 10)) THEN
            IP(J) = 0
          ELSE
            DO K = 9, 1, -1 
              IF(LogL(0) - LogL(1) < ThreShold2(J, K)) THEN
                L = K + 1
                EXIT
              END IF
            END DO
            IF(LogL(0) - LogL(1) >= ThreShold2(J, 1)) L = 1
          END IF
        END IF
      END IF
      IF(IP(J) > 0) THEN
        WRITE(Confidence(0), '(I3)') 100 - L * 5
        Confidence(0) = Trim(AdjustL(Confidence(0))) // '%'
        WRITE(10, '(3A21)') AdjustR(OffspringID(I)), &
          AdjustR(CndParentID(IP(J), J)), Confidence(0)
      END IF
    END DO
    CLOSE(10)  
  END DO
END SUBROUTINE PairwiseInference
!
SUBROUTINE Compute_PODyad_LogL(GTypeA, GTypeB, LogL, MinValue)
  USE VARIABLES
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: GTypeA(2, NumLoci), GTypeB(2, NumLoci)
  REAL(DP), INTENT(IN) :: MinValue
  REAL(DP), INTENT(OUT) :: LogL
  INTEGER :: K(2), L
  REAL(DP):: X, p(4), e1, e2, e3, e4, e5, e6, e7, d1, d2, d3, d4
  EXTERNAL :: ProbSingleGType
  LogL = 0._DP
  DO L = 1, NumLoci  
    IF(ObsNumAllele(L) == 1) CYCLE
    IF(GTypeA(1, L) <= 0 .AND. GTypeB(1, L) <= 0) CYCLE
    IF(LogL <= MinValue) RETURN
    K(1) = 0
    IF(GTypeA(1, L) <= 0) THEN
      K(1 : 2) = GTypeB(1 : 2, L)
    ELSE IF(GTypeB(1, L) <= 0) THEN
      K(1 : 2) = GTypeA(1 : 2, L)
    END IF
    IF(K(1) > 0) THEN
      CALL ProbSingleGType(L, K, X)
      X = LogL + Log(X)
      IF(X <= MinValue) THEN
        LogL = MinValue
      ELSE
        LogL = X
      END IF
      CYCLE
    END IF
    IF(LocusError(L) /= 0) THEN
      E1 = DropRate(L) / (1._DP + DropRate(L))    
      E2 = OtherErrorRate(L)
      E3 = E2 / (ObsNumAllele(L) - 1._DP)
      IF(DominantMarker(L) == 0) THEN 
        IF(GTypeA(1, L) == GTypeA(2, L)) THEN
          IF(GTypeB(1, L) == GTypeB(2, L)) THEN
            IF(GTypeA(1, L) == GTypeB(1, L)) THEN         !S1
              p(1) = ObsAlleleFre(L, GTypeA(1, L))
              p(2) = 1._DP - p(1)
              E4 = e3 - e2*e3 + e1*(-1 + e2 + e3)**2                
              X = (-1 + e2)**4*p(1)**3 + 2*(-1 + e2)**2*e4*p(1)**2*p(2) + &
                e3**4*p(2)**3 + e4*p(1)*p(2)*(e4 + 2*e3**2*p(2))
            ELSE                                          !S2
              p(1) = ObsAlleleFre(L, GTypeA(1, L))
              p(2) = ObsAlleleFre(L, GTypeB(1, L))
              E4 = -1 + e2 + e3
              E6 = 1 - e2 + e3
              E7 = -1 + 2*e1
              D2 = e3 - e1*e4
              X = -(e3 ** 4*(-1 + p(1))) + 3*d2*e3*e4 ** 2*p(1)*p(2) + &
                d2*e4 ** 3*p(1)*p(2)*(-3*e1 + e7*(p(1) + p(2))) + &
                e3 ** 2*(((-1 + e2) ** 2 + e4 ** 2*e7)*p(1) - e4 ** 2* &
                e7*p(1) ** 2 - e4*p(2)*(e4 - 2*e1*e4 + e6 + e4*e7*p(2))) 
            END IF
          ELSE IF(ANY(GTypeA(1, L) == GTypeB(1 : 2, L))) THEN  !S3
            p(1) = ObsAlleleFre(L, GTypeA(1, L))
            IF(GTypeB(1, L) == GTypeA(1, L)) THEN
              p(2) = ObsAlleleFre(L, GTypeB(2, L))
            ELSE
              p(2) = ObsAlleleFre(L, GTypeB(1, L))
            END IF   
            E4 =  -1 + e2 + e3
            E5 = -1 + 2*e1
            E6 = -1 + e2
            X = e3*(2*e3 ** 3 + e4*((-5 + 3*e1)*e3 ** 2 - &
              e1*E6 ** 2 + e3*(E6 + 2*e1*E6))*p(1) - &
              e4 ** 2*(-4*e3 + 5*e1*e3 + E6 + e1*E6)*p(1) ** 2 + &
              e4 ** 3*E5*p(1) ** 3) - e4*(2*e3 ** 3 + 2*(-2 + e1*(2 + &
              e1))*e3 ** 2*e4*p(1) + e4*E5*E6*p(1)*(e1*E6 + e4*p(1)) + &
              e3*e4*p(1)*((1 + e1*(-1 + 4*e1))*E6 - 2*e4*E5*p(1)))*p(2)     
          ELSE                                            !S4
            p(1) = ObsAlleleFre(L, GTypeA(1, L))
            P(2 : 3) = ObsAlleleFre(L, GTypeB(1 : 2, L))
            E4 = -1 + e2 + e3
            E5 = -1 + e2 - e3
            X = e3*(2*e3*(e3 ** 2 - e3*e4*p(1) + &
              e4*(-1 + e2 - (-1 + 2*e1)*e4*(-1 + p(1)))*p(1)) + &
              e4*((e4*p(1))**2*(-1 + 2*e1) - 2*e3 ** 2 + 3*e4*(e3 - &
              e1*e4)*p(1))*p(2)) + e4*(-2*e3 ** 3 - &
              3*e3*(-e3 + e1*(-1 + e2 + e3))*e4*p(1) + (-1 + &
              2*e1)*e3*e4 ** 2*p(1) ** 2 - 2*(-1 + 2*e1)* &
              e4*(e3 ** 2 + e4*(-e3 + e1*e4)*p(1))*p(2))*p(3)
          END IF
        ELSE
          IF(GTypeB(1, L) == GTypeB(2, L)) THEN
            IF(ANY(GTypeA(1 : 2, L) == GTypeB(1, L))) THEN   !S5
              p(1) = ObsAlleleFre(L, GTypeB(1, L))
              IF(GTypeA(1, L) == GTypeB(1, L)) THEN
                p(2) = ObsAlleleFre(L, GTypeA(2, L))
              ELSE
                p(2) = ObsAlleleFre(L, GTypeA(1, L))
              END IF   
              E4 =  -1 + e2 + e3
              E5 = -1 + e2
              D1 = 1 - 2*e1
              X = e3*(2*e3 ** 3 + (-5 + 3*e1)*e3 ** 2*e4*p(1) + &
                e3*e4*p(1)*(e5 + 2*e1*e5 + (4 - 5*e1)*e4*p(1)) - &
                e4*p(1)*(e1*e5 ** 2 + (1 + e1)*(-1 + e2)*e4*p(1) + &
                d1*e4 ** 2*p(1) ** 2)) - e4*(2*e3 ** 3 + 2*(-2 + &
                e1*(2 + e1))*e3 ** 2*e4*p(1) - d1*e4*e5*p(1)* &
                (e1*e5 + e4*p(1)) + e3*e4*p(1)*((1 + e1*(-1 + &
                4*e1))*e5 + 2*d1*e4*p(1)))*p(2)
            ELSE                                           !S6
              p(1) = ObsAlleleFre(L, GTypeB(1, L))
              P(2 : 3) = ObsAlleleFre(L, GTypeA(1 : 2, L))
              E4 = -1 + e2 + e3
              D3 = d2*e4*(e3 - e1*e4)*p(1)
              D4 = (-1 + 2*e1)*e4**2*p(1)**2
              X = e3*(2*e3*(e3 ** 2 + &
                2*(-1 + e2 + e3)*(-e3 + e1*(-1 + e2 + e3))* &
                p(1) - (-1 + 2*e1)*(-1 + e2 + e3) ** 2*p(1) ** 2) + (d3 + &
                d4 - 2*e3 ** 2)*e4*p(2)) - e4*(-(e3*(d3 + d4 - 2*e3 ** 2)) + &
                2*(-1 + 2*e1)*e4*(e3 ** 2 + e4*(-e3 + e1*e4)*p(1))*p(2))*p(3)
            END IF
          ELSE IF(ALL(GTypeA(1 : 2, L) == GTypeB(1 : 2, L))) THEN  !S7
            P(1 : 2) = ObsAlleleFre(L, GTypeA(1 : 2, L))
            E4 = -1 + e2 + e3
            E5 = -1 + 2*e1
            X = 4*e3 ** 4 - 7*e3 ** 3*e4*(p(1) + p(2)) + (-1 + e2)*e4 ** 3* &
              e5 ** 2*p(1)*p(2)*(p(1) + p(2)) + e3*e4 ** 2*e5*p(1)*p(2)* &
              (-4 + 4*e2 + (3 + 2*e1)*e4*(p(1) + p(2))) + &
              e3 ** 2*e4*(3*e4*p(1) ** 2 + p(2)*(-1 + e2 + 3*e4*p(2)) + &
              p(1)*(-1 + e2 + 2*(5 - 4*e1)*e4*p(2)))
          ELSE IF(ANY(GTypeA(1, L) == GTypeB(1 : 2, L)) .OR. &
                  ANY(GTypeA(2, L) == GTypeB(1 : 2, L))) THEN  !S8
            IF(ANY(GTypeA(1, L) == GTypeB(1 : 2, L))) THEN
              p(1) = ObsAlleleFre(L, GTypeA(1, L))
              p(2) = ObsAlleleFre(L, GTypeA(2, L))
            ELSE
              p(1) = ObsAlleleFre(L, GTypeA(2, L))
              p(2) = ObsAlleleFre(L, GTypeA(1, L))
            END IF
            p(3) = Sum(ObsAlleleFre(L, GTypeB(1 : 2, L))) - p(1)
            E4 = -1 + e2 + e3
            E5 = 1 - 2*e1
            X = 4*e3 ** 4 + (-1 + e2)*e4 ** 3*e5 ** 2*p(1)*p(2)*p(3) - &
              e3 ** 3*e4*(7*p(1) + 4*(p(2) + p(3))) - e3*e4 ** 2*e5* &
              p(1)*((-1 + e2 + 2*e4*p(1))*p(3) + p(2)*(-1 + e2 + 2*e4*p(1) + &
              (3 + 2*e1)*e4*p(3))) + e3 ** 2*e4*(3*e4*p(1) ** 2 + &
              3*e4*p(2)*p(3) + p(1)*(-1 + e2 - 6*(-1 + e1)*e4*(p(2) + p(3))))
          ELSE       !S9
            P(1 : 2) = ObsAlleleFre(L, GTypeA(1 : 2, L))
            P(3 : 4) = ObsAlleleFre(L, GTypeB(1 : 2, L))
            E4 = -1 + e2 + e3
            D2 = -4*e3**2 + 2*(-1 + 2*e1)*e4**2*p(1)*p(2) + 3*e3*e4*(p(1) + p(2))
            X =  e3*(4*e3 ** 2*(e3 - e4*p(1)) - 4*e3*e4*(e3 + (-1 + &
              2*e1)*e4*p(1))*p(2) + d2*e4*p(3) + e4*(d2 + 2*(-1 + &
              2*e1)*e4*(-2*e3 + e4*(p(1) + p(2)))*p(3))*p(4))
          END IF
        END IF
      ELSE     !Dominant marker
        p(1) = ObsAlleleFre(L, 1)
        IF(GTypeA(1, L) == 2 .AND. GTypeB(1, L) == 2) THEN
          E4 = 1 - 2*e2
          X = ((-1 + e2)**2 - E4*p(1))*(-1 + e2 + E4*p(1))**2
        ELSE IF(GTypeA(1, L) == 2 .OR. GTypeB(1, L) == 2) THEN
          X = -((1 - e2 + (-1 + 2*e2)*p(1))**2*((-2 + e2)*e2 + (-1 + 2*e2)*p(1)))
        ELSE
          X = (-2 + e2)**2*e2**2 - (-1 + 2*e2)*(1 + e2*(4 + e2*(-7 + &
            2*e2)))*p(1) + (1 - 2*e2)**2*(1 + (-4 + e2)*e2)*p(1)**2 + &
            ((-1 + 2*e2)*p(1))**3
        END IF
      END IF      
    ELSE
      IF(DominantMarker(L) == 0) THEN 
        IF(GTypeA(1, L) == GTypeA(2, L)) THEN
          IF(GTypeB(1, L) == GTypeB(2, L)) THEN
            IF(GTypeA(1, L) == GTypeB(1, L)) THEN              !S1
              X = ObsAlleleFre(L, GTypeA(1, L))**3
            ELSE                                               !S2
              X = 0._DP  
            END IF
          ELSE IF(ANY(GTypeA(1, L) == GTypeB(1 : 2, L))) THEN  !S3
            IF(GTypeB(1, L) == GTypeA(1, L)) THEN
              X = ObsAlleleFre(L, GTypeA(1, L))**2 * ObsAlleleFre(L, GTypeB(2, L))
            ELSE
              X = ObsAlleleFre(L, GTypeA(1, L))**2 * ObsAlleleFre(L, GTypeB(1, L))
            END IF   
          ELSE                                                 !S4
            X = 0._DP
          END IF
        ELSE
          IF(GTypeB(1, L) == GTypeB(2, L)) THEN
            IF(ANY(GTypeA(1 : 2, L) == GTypeB(1, L))) THEN     !S5
              IF(GTypeA(1, L) == GTypeB(1, L)) THEN
                X = ObsAlleleFre(L, GTypeB(1, L))**2 * ObsAlleleFre(L, GTypeA(2, L))
              ELSE
                X = ObsAlleleFre(L, GTypeB(1, L))**2 * ObsAlleleFre(L, GTypeA(1, L))
              END IF   
            ELSE                                               !S6
              X = 0._DP
            END IF
          ELSE IF(ALL(GTypeA(1 : 2, L) == GTypeB(1 : 2, L))) THEN  !S7
            P(1 : 2) = ObsAlleleFre(L, GTypeA(1 : 2, L))
            X = p(1)*p(2)*(p(1) + p(2))
          ELSE IF(ANY(GTypeA(1, L) == GTypeB(1 : 2, L)) .OR. &
                  ANY(GTypeA(2, L) == GTypeB(1 : 2, L))) THEN  !S8
            IF(ANY(GTypeA(1, L) == GTypeB(1 : 2, L))) THEN
              p(1) = ObsAlleleFre(L, GTypeA(1, L))
              p(2) = ObsAlleleFre(L, GTypeA(2, L))
            ELSE
              p(1) = ObsAlleleFre(L, GTypeA(2, L))
              p(2) = ObsAlleleFre(L, GTypeA(1, L))
            END IF
            p(3) = Sum(ObsAlleleFre(L, GTypeB(1 : 2, L))) - p(1)
            X = p(1)*p(2)*p(3)
          ELSE       !S9
            X = 0._DP
          END IF
        END IF
      ELSE     !Dominant marker
        p(1) = ObsAlleleFre(L, 1)     
        IF(GTypeA(1, L) == 2 .AND. GTypeB(1, L) == 2) THEN
          X = (1 - p(1))**3
        ELSE IF(GTypeA(1, L) == 2 .OR. GTypeB(1, L) == 2) THEN
          X = (1 - p(1))**2*p(1)
        ELSE
          X = p(1) + p(1)**2 - p(1)**3
        END IF
      END IF
    END IF
    IF(X > Small) THEN
      X = LogL + Log(X)
      IF(X <= MinValue) THEN
        LogL = MinValue
      ELSE
        LogL = X
      END IF
    ELSE
      LogL = MinValue
    END IF     
  END DO     
END SUBROUTINE Compute_PODyad_LogL 
!
SUBROUTINE DetermineDeltaValue
!C Determine Delta value by simulation in pairwise parentage inference
  USE SimuVariables
  IMPLICIT NONE
  INTEGER(I4B), Parameter :: NumRepeats = 10000
  INTEGER(I4B) :: I, J, K, L, N, II, ISex, IndivPrntCopy(2), GType(2), &
    NumSimuCandidate(2)
  INTEGER(I4B), Allocatable :: SimuOffGType(:, :), SimPrntGType(:, :, :), &
    OffGtypeCopy(:, :), CndGtypeCopy(:, :, :), OffGType2Copy(:), Idx(:)
  REAL(DP) :: X, Delta(NumRepeats), Delta2(NumRepeats), Best(2), Best2(2)
  REAL(DP) :: LOD
  REAL(DP), ALLOCATABLE :: LogL(:), LogL2(:)
  REAL(DP), ALLOCATABLE :: Indiv1Locus1LogL(:, :, :)
  Logical :: TrueParentQ(NumRepeats), SampledQ, QuickDomQ
  REAL :: R
  EXTERNAL :: SimuGenotype, SimuDomGenotype, SimuGenotypeChange, Indexx, &
    Compute_PODyad_LogL
  Allocate(SimuOffGType(2, NumLoci), SimPrntGType(2, NumLoci, 2), &
    MissingRate(NumLoci), OffGtypeCopy(2, NumLoci), &
    CndGtypeCopy(2, NumLoci, 2), Idx(NumRepeats), &
    OffGType2Copy(NumLoci), LogL(3), LogL2(3), &
    Indiv1Locus1LogL(MaxNumGenotype, NumLoci, 2))
  IF(Sum(DominantMarker(:)) == NumLoci) THEN
    L = 3
    QuickDomQ = .TRUE.
  ELSE
    L = MaxObsNumallele
    QuickDomQ = .FALSE.
  END IF
  Allocate(CumulativeAlleleFre(L, NumLoci))
  Indiv1Locus1LogL = 1.0_DP
!C Calculate cumulative allele freq. & missing rate
!    for each locus used in simulation
  DO L = 1, NumLoci
    IF(.NOT. QuickDomQ) THEN
      CumulativeAlleleFre(1, L) = ObsAlleleFre(L, 1)
      DO I = 2, ObsNumAllele(L) - 1
        CumulativeAlleleFre(I, L) = CumulativeAlleleFre(I - 1, L) + &
          ObsAlleleFre(L, I)
      END DO
      CumulativeAlleleFre(ObsNumAllele(L), L) = 2.0
    ELSE
      CumulativeAlleleFre(1, L) = ObsAlleleFre(L, 1) ** 2
      CumulativeAlleleFre(2, L) = CumulativeAlleleFre(1, L) + &
        2._DP * ObsAlleleFre(L, 1) * ObsAlleleFre(L, 2)
      CumulativeAlleleFre(3, L) = 2.0
    END IF
    K = 0
    DO I = 1, SmplSize
      IF(IndivGType(1, L, I) < 1) K = K + 1
    END DO
    DO J = 1, 2
      DO I = 1, NumCandidate(J)
        IF(CndGType(1, L, I, J) < 1) K = K + 1
      END DO
    END DO
    MissingRate(L) = Dble(K) / (SmplSize + Sum(NumCandidate(1 : 2)))
  END DO
  AnyGtypeMissingQ = ANY(MissingRate(1 : NumLoci) > Small)
!C Initialise variables for likelihood calculation
  OffGtypeCopy(1 : 2, 1 : NumLoci) = IndivGType(1 : 2, 1 : NumLoci, 1)
  OffGType2Copy(1 : NumLoci) = IndivGType2(1 : NumLoci, 1)
  DO I = 1, 2
    IF(NumCandidate(I) < 1) CYCLE
    CndGtypeCopy(:, :, I) = CndGtype(:, :, 1, I)
    IndivPrntCopy(I) = IndivParent(1, I)
  END DO
!C Simulating
  WRITE(*, *)'Estimating parentage with pairwise approach'
  DO ISex = 1, 2
    X = Sum(AgeSize(MinParentAge(Isex) : , ISex))
    NumSimuCandidate(ISex) = NINT(X / PropPrntSampled(ISex))
  END DO  
  DO ISex = 1, 2
    IF(NumCandidate(ISex) < 1) CYCLE
    IF(ISex == 2 .AND. NumCandidate(2) == NumCandidate(1) .AND. &
       NumSimuCandidate(2) == NumSimuCandidate(1) .AND. &
       (ALL(NumKnownPrnt(1 : 2) > 0) .OR. ALL(NumKnownPrnt(1 : 2) == 0)))THEN
      ThreShold(2, :) = ThreShold(1, :)
      IF(NumKnownPrnt(2) > 0) ThreShold2(2, :) = ThreShold2(1, :)
      EXIT
    END IF
    DO N = 1, NumRepeats
      IF(MOD(N, 100) == 0) WRITE(*, Merge('(A,$)', '(A)  ', MOD(N, 10000) /= 0)) '.'
!C Generate genotypes
      IF(.NOT. QuickDomQ) THEN
        DO I = 1, 2
          CALL SimuGenotype(SimPrntGType(:, :, I))
        END DO
      ELSE
        DO I = 1, 2
          CALL SimuDomGenotype(SimPrntGType(:, :, I))
        END DO
      END IF
      DO L = 1, NumLoci
        DO I = 1, 2
          CALL Random_Number(R)
          SimuOffGType(I, L) = SimPrntGType(NINT(R) + 1, L, I)
        END DO
      END DO
      CALL SimuGenotypeChange(SimuOffGType(:, :))
      DO I = 1, 2
        CALL SimuGenotypeChange(SimPrntGType(:, :, I))
      END DO
      DO L = 1, NumLoci
        IndivGType(1 : 2, L, 1) = SimuOffGType(1 : 2, L)
        IF(SimuOffGType(1, L) < 1)THEN
          IndivGType2(L, 1) = 0
        ELSE
          IF(DominantMarker(L) == 1) THEN
            IndivGType2(L, 1) = SimuOffGType(1, L)
          ELSE
            IndivGType2(L, 1) = GenotypeID2(SimuOffGType(1, L), SimuOffGType(2, L))
          END IF
        END IF
      END DO
      IndivParent(1, :) = 0
      LogL(1) = 0._DP
      DO L = 1, NumLoci
        IF(IndivGType2(L, 1) < 1) CYCLE
        IF(Indiv1Locus1LogL(IndivGType2(L, 1), L, 2) > 0.1_DP) THEN
          GType(1 : 2) = IndivGType(1 : 2, L, 1)
          CALL ProbSingleGType(L, GType, X)
          Indiv1Locus1LogL(IndivGType2(L, 1), L, 2) = Log(X)
        END IF
        LogL(1) = LogL(1) + Indiv1Locus1LogL(IndivGType2(L, 1), L, 2)
      END DO
      Best = 0._DP
      K = 0
      IF(NumKnownPrnt(3 - ISex) > 0)THEN
        CndGtype(:, :, 1, 3 - ISex) = SimPrntGType(:, :, 3 - ISex)  !USED LATER
        CALL Compute_PODyad_LogL(IndivGType(:, :, 1), SimPrntGType(:, :, 3 - ISex), &
          LogL2(1), TinyValue)
        Best2 = 0._DP
        L = 0
      END IF
      CALL Random_Number(R)
      SampledQ = (R <= PropPrntSampled(ISex))
      DO I = 1, NumSimuCandidate(ISex)
        IF(I > 1 .OR. (.NOT. SampledQ)) THEN
          IF(.NOT. QuickDomQ) THEN          
            CALL SimuGenotype(CndGtype(:, :, 1, ISex))
          ELSE
            CALL SimuDomGenotype(CndGtype(:, :, 1, ISex))
          END IF
          CALL SimuGenotypeChange(CndGtype(:, :, 1, ISex))
        ELSE
          CndGtype(:, :, 1, ISex) = SimPrntGType(:, :, ISex)
        END IF
        LogL(2) = 0._DP
        DO J = 1, NumLoci
          IF(CndGtype(1, J, 1, ISex) < 1) CYCLE
          IF(DominantMarker(J) == 1) THEN
            II = CndGtype(1, J, 1, ISex)
          ELSE
            II = GenotypeID2(CndGtype(1, J, 1, ISex), CndGtype(2, J, 1, ISex))
          END IF
          IF(Indiv1Locus1LogL(II, J, 2) > 0.1_DP) THEN
            GType(1 : 2) = CndGtype(1 : 2, J, 1, ISex)
            CALL ProbSingleGType(J, GType, X)
            Indiv1Locus1LogL(II, J, 2) = Log(X)
          END IF
          LogL(2) = LogL(2) + Indiv1Locus1LogL(II, J, 2)
        END DO
        X = Sum(LogL(1 : 2))
        CALL Compute_PODyad_LogL(IndivGType(:, :, 1), &
          CndGtype(:, :, 1, ISex), LogL(3), X + Best(2))
        LOD = LogL(3) - X
        IF(LOD > Best(1))THEN
          Best(2) = Best(1)
          Best(1) = LOD
          K = I
        ELSE IF(LOD > Best(2))THEN
          Best(2) = LOD
        END IF
      END DO
      Delta(N) = Best(1) - Best(2)
      IF(SampledQ) THEN
        TrueParentQ(N) = (K == 1)
      ELSE
        TrueParentQ(N) = .FALSE.
      END IF
      IF(NumKnownPrnt(3 - ISex) > 0) Delta2(N) = Best2(1) - Best2(2)
    END DO
    CALL Indexx(NumRepeats, Delta, Idx)
    DO N = 95, 50, -5
      X = Dble(N) * 0.01_DP
      J = 0
      K = 0
      DO I = NumRepeats, 1, -1
        IF(.NOT. TrueParentQ(Idx(I))) CYCLE
        J = J + 1
        IF(DBLE(J) / (NumRepeats - I + 1) >= X) THEN
          ThreShold(ISex, 20 - N / 5) = Delta(Idx(I))
          K = 1
        ELSE
          IF(K == 1) EXIT
        END IF
      END DO
      IF(K /= 1) ThreShold(ISex, 20 - N / 5) = 999.9
    END DO
    IF(NumKnownPrnt(3 - ISex) > 0) THEN
      CALL Indexx(NumRepeats, Delta2, Idx)
      DO N = 95, 50, -5
        X = Dble(N) * 0.01_DP
        J = 0
        K = 0
        DO I = NumRepeats, 1, -1
          IF(.NOT. TrueParentQ(Idx(I))) CYCLE
          J = J + 1
          IF(DBLE(J) / (NumRepeats - I + 1) >= X) THEN
            ThreShold2(ISex, 20 - N / 5) = Delta2(Idx(I))
            K = 1
          ELSE
            IF(K == 1) EXIT
          END IF
        END DO
        IF(K /= 1) ThreShold2(ISex, 20 - N / 5) = 999.9
      END DO
    END IF  
  END DO
!C Copy back the original offspring & candidate genotypes
  IndivGType(1 : 2, 1 : NumLoci, 1) = OffGtypeCopy(1 : 2, 1 : NumLoci)
  IndivGType2(1 : NumLoci, 1) = OffGType2Copy(1 : NumLoci)
  DO I = 1, 2
    CndGtype(:, :, 1, I) = CndGtypeCopy(:, :, I)
    IndivParent(1, I) = IndivPrntCopy(I)
  END DO
  WRITE(*,*)
END SUBROUTINE DetermineDeltaValue
!
SUBROUTINE SimuGenotype(Gtype)
!C Simulate multi-locus genotypes used in pairwise parentage inference
  USE SimuVariables
  IMPLICIT NONE
  INTEGER(I4B), INTENT(OUT) :: Gtype(2, NumLoci)
  INTEGER(I4B) :: I, J, K, L
  REAL :: UniformR(2, NumLoci)     
  CALL Random_Number(UniformR)
  DO L = 1, NumLoci
    IF(UniformR(1, L) > UniformR(2, L)) THEN
      I = 2
      K = 1
    ELSE
      I = 1
      K = 2
    END IF
    DO J = 1, ObsNumAllele(L)
      IF(UniformR(I, L) <= CumulativeAlleleFre(J, L)) THEN
        GType(1, L) = J 
        EXIT
      END IF
    END DO
    DO J = J, ObsNumAllele(L)
      IF(UniformR(K, L) <= CumulativeAlleleFre(J, L)) THEN
        GType(2, L) = J
        EXIT
      END IF
    END DO     
  END DO
END SUBROUTINE SimuGenotype
!
SUBROUTINE SimuDomGenotype(Gtype)
!C Simulate multi-locus genotypes used in pairwise parentage inference
  USE SimuVariables
  IMPLICIT NONE
  INTEGER(I4B), INTENT(OUT) :: Gtype(2, NumLoci)
  INTEGER(I4B) :: L
  REAL :: R(NumLoci)
  CALL Random_Number(R)  
  DO L = 1, NumLoci
    IF(R(L) <= CumulativeAlleleFre(1, L))THEN
      GType(1 : 2, L) = 1
    ELSE IF(R(L) <= CumulativeAlleleFre(2, L))THEN
      GType(1 : 2, L) = (/1, 2/)
    ELSE
      GType(1 : 2, L) = 2
    END IF
  END DO
END SUBROUTINE SimuDomGenotype
!
SUBROUTINE SimuGenotypeChange(Gtype)
!C Change simulated genotypes for typing errors and missing
  USE SimuVariables
  IMPLICIT NONE
  INTEGER(I4B), INTENT(INOUT) :: Gtype(2, NumLoci)
  INTEGER(I4B) :: I, L
  REAL :: R
  IF(.NOT. MarkerErrorFreeQ) THEN
    DO I = 1, NumLoci
      IF(DropRate(I) <= Small) CYCLE
      IF(Gtype(1, I) == Gtype(2, I)) CYCLE
      CALL Random_Number(R)
      IF(R < DropRate(I)) THEN
        CALL Random_Number(R)
        L = NINT(R) + 1
        GType(3 - L, I) = GType(L, I)
      END IF
    END DO
    DO I = 1, NumLoci
      IF(OtherErrorRate(I) <= Small) CYCLE
      DO L = 1, 2
        CALL Random_Number(R)
        IF(R < OtherErrorRate(I)) THEN
          CALL Random_Number(R)
          GType(L, I) = NINT(R * (ObsNumAllele(I) - 1)) + 1
        END IF
      END DO
    END DO
  END IF
  DO I = 1, NumLoci
    IF(DominantMarker(I) == 1) THEN
      IF(Gtype(1, I) /= Gtype(2, I)) Gtype(1 : 2, I) = 1
    ELSE IF(Gtype(1, I) > Gtype(2, I)) THEN
      Gtype(1 : 2, I) = Gtype(2 : 1 : -1, I)
    END IF
  END DO
  IF(AnyGtypeMissingQ) THEN
    DO I = 1, NumLoci
      IF(MissingRate(I) > Small)THEN
        CALL Random_Number(R)
        IF(R < MissingRate(I)) GType(:, I) = 0
      END IF
    END DO
  END IF
END SUBROUTINE SimuGenotypeChange
!
SUBROUTINE ReadData
  USE VARIABLES
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L, M, N, K1, K2, IOS, NumLines
  INTEGER(I4B), Allocatable :: Genotype(:), Idx(:) 
  CHARACTER(Len = 20) :: CH(2)
  CHARACTER(Len = 20), Allocatable :: C(:)
  EXTERNAL :: StopOnError, indexCh
  ErrorMessage = ''  
  ErrorMessage(2) = 'AgeSexGtype.dat'
  NumLines = 0
  OPEN(UNIT=10,FILE='AgeSexGtype.dat',IOSTAT=IOS,ERR=1,STATUS='OLD')
  ErrorMessage(2) = ''
  ErrorMessage(1) = 'Error in reading line 1, Dataset Name'  
  READ(10,*,IOSTAT=IOS,ERR=9,End=9)DataSetName
  ErrorMessage(1) = 'Error in reading line 2, Output File Name'  
  READ(10,*,IOSTAT=IOS,ERR=9,End=9)FName
  FName = Trim(AdjustL(FName))
  ErrorMessage(1) = 'Error in reading line 3, Offspring Sample Size'    
  READ(10,*,IOSTAT=IOS,ERR=9,End=9)SmplSize
  IF(SmplSize < 1) THEN
    WRITE(ErrorMessage(1),'(A)')&
      'The sample size of offspring should be greater than 0!'
    GOTO 9
  END IF
  ErrorMessage(1) = 'Error in reading line 4, Maximal age for males & females'      
  READ(10, *,IOSTAT=IOS,ERR=9,End=9)NumAgeClass(1 : 2) 
  ErrorMessage(1) = '' 
  IF(NumAgeClass(1) < 1) THEN
    ErrorMessage(1) = 'The maximal age for males is too small!'
  ELSE IF(NumAgeClass(1) > 50) THEN
    ErrorMessage(1) = 'The maximal age for males is too large!'
  ELSE IF(NumAgeClass(2) < 1) THEN
    ErrorMessage(1) = 'The maximal age for females is too small!'
  ELSE IF(NumAgeClass(2) > 50) THEN
    ErrorMessage(1) = 'The maximal age for females is too large!'
  END IF 
  IF(ErrorMessage(1) /= '') GOTO 9
  NumAgeClass(0) = NumAgeClass(1) + NumAgeClass(2)
  ErrorMessage(1) = 'Error in reading line 5, ' // &
    'Minimal reproductive age for males & females'      
  READ(10, *,IOSTAT=IOS,ERR=9,End=9) MinParentAge(1 : 2)
  ErrorMessage(1) = '' 
  IF(MinParentAge(1) < 1) THEN
    ErrorMessage(1) = 'The minimal reproductive age for males is too small!'
  ELSE IF(MinParentAge(1) > NumAgeClass(1)) THEN
    ErrorMessage(1) = 'The minimal reproductive age for males is too large!'
  ELSE IF(MinParentAge(2) < 1) THEN
    ErrorMessage(1) = 'The minimal reproductive age for females is too small!'
  ELSE IF(MinParentAge(2) > NumAgeClass(2)) THEN
    ErrorMessage(1) = 'The minimal reproductive age for females is too large!'
  END IF 
  IF(ErrorMessage(1) /= '') GOTO 9  
  ErrorMessage(1) = 'Error in reading line 6, assumed family size distribution!' 
  READ(10, *,IOSTAT=IOS,ERR=9,End=9)FamilySizeDist
  IF(FamilySizeDist /= 0 .AND. FamilySizeDist /= 1) THEN
    ErrorMessage(1) = 'The indicator for assumed family size distribution' &
      // ' must be 1/0 for Poisson/Other distribution!'
    GOTO 9
  END IF
  ErrorMessage(1) = 'Error in reading line 7, sampling proportion' // &
    ' known/unknown for each age class!' 
  READ(10, *,IOSTAT=IOS,ERR=9,End=9) SamplingPrbKnown
  IF(SamplingPrbKnown /= 0 .AND. SamplingPrbKnown /= 1) THEN
    ErrorMessage(1) = 'The indicator for sampling proportion' &
      // ' must be 1/0 for known/unknown!'
    GOTO 9
  END IF
  
  IF(SamplingPrbKnown == 1) THEN
    ErrorMessage(1) = 'Error in reading known sampling proportions of each age class'
    Allocate(SamplingPrbPrior(2))
    DO K = 1, 2
      Allocate(SamplingPrbPrior(K)%R(NumAgeClass(K)))
      READ(10, *,IOSTAT=IOS,ERR=9,End=9)(SamplingPrbPrior(K)%R(J), J = 1, NumAgeClass(K))
      DO J = 1, NumAgeClass(K)
        IF(SamplingPrbPrior(K)%R(J) > 1._DP .OR. &
           SamplingPrbPrior(K)%R(J) < 0._DP) THEN
          ErrorMessage(1) = 'Invalid values for known sampling proportions!'
          GOTO 9
        END IF
      END DO
    END DO
    NumLines = 9
  ELSE
    NumLines = 7
  END IF
  NumLines = NumLines + 1
  WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', NumLines, &
    ', number of bootstrapping samples!'
  READ(10, *,IOSTAT=IOS,ERR=9,End=9)NumCI95Sample
  ErrorMessage(1) = ''
  IF(NumCI95Sample < 0) THEN
    ErrorMessage(1) = 'The number of bootstrapping samples must be >=0!'
  ELSE IF(NumCI95Sample > 100000) THEN
    ErrorMessage(1) = 'The number of bootstrapping samples is too big!'
  END IF
  IF(ErrorMessage(1) /= '') GOTO 9
  NumLines = NumLines + 1
  WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &
    NumLines, ', Number of Loci'      
  READ(10,*,IOSTAT=IOS,ERR=9,End=9)NumLoci
  IF(NumLoci < 1) THEN
    WRITE(ErrorMessage(1),'(A)')&
      'The number of loci should be greater than 0!'
    GOTO 9
  END IF
  NumLines = NumLines + 1
  WRITE(ErrorMessage, '(A,I4, A)') 'Error in reading line', &
    NumLines, ', Confidence level for parentage assignment'      
  READ(10,*,IOSTAT=IOS,ERR=9,End=9)I
  IF(I < 80 .AND. I > 95 .AND. Mod(I, 5) /= 0) THEN
    ErrorMessage = 'The confidence level for parentage' // &
      ' assignment must be one of {80, 85, 90, 95}!'
    GOTO 9
  END IF  
  NumLines = NumLines + 1
  WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &
    Numlines, ', Marker IDs'   
  ALLOCATE(DropRate(NumLoci), OtherErrorRate(NumLoci), &
    DominantMarker(NumLoci), MarkerID(NumLoci))
  READ(10,*,IOSTAT=IOS,Err=9,End=9)(MarkerID(I), I = 1, NumLoci)
  DO I = 1, NumLoci
    MarkerID(I) = Trim(AdjustR(MarkerID(I)))
  END DO 
  IF(Allocated(SortCharacter)) Deallocate(SortCharacter)   
  ALLOCATE(Idx(NumLoci), SortCharacter(NumLoci))
  DO I = 1, NumLoci 
    SortCharacter(I) = MarkerID(I)
  END DO
  CALL indexCh(NumLoci, Len(SortCharacter(1)), idx)  
  DO I = 2, NumLoci
    IF(MarkerID(Idx(I)) == MarkerID(Idx(I - 1))) THEN
      WRITE(ErrorMessage(1),'(A,I8,A,A,A,I8)') 'Marker', Idx(I), &
        ' has the same ID, ', Trim(MarkerID(Idx(I))), ', as Marker', Idx(I - 1)
      GOTO 9
    END IF
  END DO
  NumLines = NumLines + 1
  WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &
    Numlines, ', indicators for marker types per locus'   
  READ(10,*,IOSTAT=IOS,Err=9,End=9)(DominantMarker(I), I = 1, NumLoci)
  IF(ANY(DominantMarker(:) > 1) .OR. ANY(DominantMarker(:) < 0)) THEN
    WRITE(ErrorMessage(1),'(A)')&
      'Marker type should be either 1 (dominant) or 0 (codominant)!'
    GOTO 9
  END IF
  NumLines = NumLines + 1
  WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &
    Numlines, ', Dropout Rate per locus'  
  READ(10,*,IOSTAT=IOS,Err=9,End=9)DropRate(1 : NumLoci)
         !Drop rate for each locus
  NumLines = NumLines + 1
  WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &
    Numlines, ', Other Error Rate per locus'          
  READ(10,*,IOSTAT=IOS,Err=9,End=9)OtherErrorRate(1 : NumLoci)
         !OtherErrorRate for each locus
  IF(ANY(DropRate(:) <= -Small) .OR. ANY(OtherErrorRate(:) <= -Small) .OR. &
    ANY(DropRate(:) > 1._DP) .OR. ANY(OtherErrorRate(:) > 1._DP)) THEN
    WRITE(ErrorMessage(1),'(A)')&
      'Marker error rate is out of the range [0, 1]!'
    GOTO 9
  END IF
  ALLOCATE(IndivGType(2, NumLoci, SmplSize), Genotype(NumLoci), &
    OffspringID(SmplSize), IndivSex(SmplSize), IndivAge(SmplSize), STAT = I)
  IF(I /= 0) Call StopOnError(0)
  DO I = 1, NumLoci
    Genotype(I) = 2 - DominantMarker(I)
  END DO
  DO I = 1, SmplSize
    NumLines = NumLines + 1
    WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &
      Numlines, ', Data or data format error in offspring ID/Sex/Age/genotypes'   
    READ(10,*,IOSTAT=IOS,Err=9,End=9) OffspringID(I), IndivSex(I), IndivAge(I), &
      (IndivGType(1 : Genotype(J), J, I), J = 1, NumLoci)
  END DO
  IF(Allocated(Idx)) Deallocate(Idx)
  IF(Allocated(SortCharacter)) Deallocate(SortCharacter)
  ALLOCATE(Idx(SmplSize), SortCharacter(SmplSize))
  DO I = 1, SmplSize 
    SortCharacter(I) = OffspringID(I)
  END DO
  CALL indexCh(SmplSize, Len(SortCharacter(1)), idx)  
  DO I = 2, SmplSize
    IF(OffspringID(Idx(I)) == OffspringID(Idx(I - 1)))THEN
      WRITE(ErrorMessage(1),'(A,I8,A,A,A,I8)')&
        'Offspring', Idx(I), ' has the same ID, ', Trim(OffspringID(Idx(I))), &
        ', as offspring', Idx(I - 1)
      GOTO 9
    END IF
  END DO
  IF(Allocated(Idx)) Deallocate(Idx)
  Allocate(Idx(NumLoci))
  DO J = 1, NumLoci
    IF(DominantMarker(J) == 1) THEN
      DO I = 1, SmplSize
        IndivGType(2, J, I) = IndivGType(1, J, I)
      END DO
    END IF
  END DO
  NumLines = NumLines + 1
  WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &
    NumLines, ', Proportions of parents sampled'   
  READ(10,*,IOSTAT=IOS,Err=9,End=9)PropPrntSampled(1 : 2)
          !Prop. of fathers & mothers sampled in candidates
  IF(Any(PropPrntSampled(:) < 1.E-3_DP) .AND. Any(NumCandidate(1 : 2) > 0))THEN
    WRITE(ErrorMessage(1),'(A,I5,A,A,A)') &
      'Probability of a parent included in candidates should be >= 0.001!'
    GOTO 9
  END IF
  NumCandidate(1 : 2) = SmplSize
  ALLOCATE(CndGType(2, NumLoci, SmplSize, 2), CndParentID(0 : SmplSize, 2))
  DO J = 1, 2
    DO I = 1, SmplSize
      CndParentID(I, J) = OffspringID(I)
      DO L = 1, NumLoci
        CndGType(1 : 2, L, I, J) = IndivGType(1 : 2, L, I)
      END DO
    END DO
  END DO  
  I = Maxval(NumAgeClass(1 : 2))
  Allocate(AgeSize(0 : I, 2))   
  AgeSize(:, :) = 0
  DO I = 1, SmplSize
    AgeSize(IndivAge(I), IndivSex(I)) = AgeSize(IndivAge(I), IndivSex(I)) + 1      
  END DO
  ALLOCATE(IndivParent(SmplSize, 2))
  IndivParent = -1
  DO K = 1, 2
    NumLines = NumLines + 1
    IF(K == 1) THEN
      WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &
        Numlines, ', known paternity'
    ELSE
      WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &
        Numlines, ', known maternity'
    END IF        
    READ(10, *, IOSTAT=IOS, Err=9, End=9) J
      !Number of offspring with known father/mother
    DO I = 1, J
      NumLines = NumLines + 1
      WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &
        Numlines, ', known offspring-parent IDs'
      READ(10,*,IOSTAT=IOS,Err=9,End=9) CH(1 : 2)
      K1 = 0
      DO N = 1, SmplSize
        IF(CH(1) == OffspringID(N)) THEN
          K1 = N
          EXIT
        END IF
      END DO
      IF(K1 == 0) THEN
        WRITE(ErrorMessage(1), *)'Offspring ID, ', &
          Trim(CH(1)), ', not found in the offspring pool!'
        GOTO 9
      END IF
      L = 0
      DO N = 1, NumCandidate(K)
        IF(CH(2) == CndParentID(N, K)) THEN
          L = N
          EXIT
        END IF
      END DO
      IF(L == 0) THEN
        WRITE(ErrorMessage(1), *)'Candidate ID, ', &
          Trim(CH(2)), ', not found in the Candidate pool!'
        GOTO 9
      END IF
      IndivParent(K1, K) = L
    END DO
  END DO
  IF(Allocated(C)) Deallocate(C)
  ALLOCATE(C(SmplSize))
  IF(Allocated(Idx)) Deallocate(Idx)
  Allocate(Idx(SmplSize))
  ALLOCATE(ExcludedParent(SmplSize, 2))
  ExcludedParent(:, :)%N = 0
  DO I = 1, 2   !Read excluded male & female candidates
    NumLines = NumLines + 1
    IF(I == 1) THEN
      WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', & 
        NumLines, ', Known excluded paternity'
    ELSE
      WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &  
        NumLines, ', Known excluded maternity'
    END IF     
    READ(10,*,IOSTAT=IOS,Err=9,End=9) K
      !Number of offspring with known excluded candidates as father/mother
    IF(K > 0) THEN
      IF(Allocated(Idx)) Deallocate(Idx)
      J = NumCandidate(I)
      Allocate(Idx(J + J))
      SortCharacter(1 : J) = CndParentID(1 : J, I)      
      CALL indexCh(J, Len(SortCharacter(1)), Idx(1 : J))
    END IF
    K2 = 0
    DO J = 1, K
      NumLines = NumLines + 1
      WRITE(ErrorMessage(1), '(A,I4, A)') 'Error in reading line', &  
        NumLines, ', Known excluded paternity/maternity'
      READ(10,*,IOSTAT=IOS,Err=9,End=9) CH(1), L, C(1 : L)
             !OffspringID, # excluded parent, Excluded parent IDs
      M = 0
      DO N = 1, SmplSize
        IF(CH(1) == OffspringID(N)) THEN
          M = 1
          EXIT
        END IF
      END DO
      IF(M == 0)THEN
        WRITE(ErrorMessage(1), *)'Offspring ID, ', &
          Trim(CH(1)), ', not found in the offspring pool!'
        GOTO 9
      END IF
      ExcludedParent(N, I)%N = L
      IF(L <= 0) CYCLE
      ALLOCATE(ExcludedParent(N, I)%List(L))
      IF(K2 > 0 .AND. L == ExcludedParent(K2, I)%N) THEN
        IOS = 1        
        DO M = 1, L
          IF(C(M) /= SortCharacter(M)) THEN
            IOS = 0
            EXIT
          END IF
        END DO
      ELSE
        IOS = 0
      END IF
      IF(IOS == 0) THEN
        SortCharacter(1 : L) = C(1 : L)      
        M = NumCandidate(I)
        CALL indexCh(L, Len(SortCharacter(1)), Idx(M + 1 : M + L))
        IOS = 0
        DO M = 1, L
          ExcludedParent(N, I)%List(M) = 0
          DO K1 = IOS + 1, NumCandidate(I)
            IF(C(Idx(NumCandidate(I) + M)) == CndParentID(Idx(K1), I)) THEN
              ExcludedParent(N, I)%List(M) = Idx(K1)
              IOS = K1
              EXIT
            END IF
          END DO
          IF(ExcludedParent(N, I)%List(M) == 0)THEN
            WRITE(ErrorMessage(1), *)'Candidate ID, ', &
              Trim(C(M)), ', not found in the Candidate pool!'
            GOTO 9
          END IF
        END DO
        K2 = N
      ELSE
        DO M = 1, L
          ExcludedParent(N, I)%List(M) = ExcludedParent(K2, I)%List(M)
        END DO
      END IF
    END DO  
  END DO
  CLOSE(10)
  ErrorMessage(1) = ''
  RETURN
1 WRITE(ErrorMessage(1),'(A)') 'Input file ' // &
    Trim(AdjustL(ErrorMessage(2))) // ' does not exist!'
  ErrorMessage(2) = ''
  GOTO 7
9 WRITE(ErrorMessage(2),'(A)') &
    'Errors in DATA. Insufficient data or incorrect format.'
  WRITE(ErrorMessage(3),'(A)') &
    'Please check DATA and format and then re-run the program'
7 Call StopOnError(1)
END SUBROUTINE ReadData
!
SUBROUTINE StopOnError(IFlag)
  USE VARIABLES
  IMPLICIT NONE
  REAL :: DelayTime(2)
  INTEGER :: I, IFlag
  Call CPU_TIME(DelayTime(1))
  Open(10, File = 'ErrorMessage')
  IF(IFlag == 0) THEN
    WRITE(*, *) 'Program stops due to insufficient memory!'
    WRITE(10, *) 'Program stops due to insufficient memory!'
  ELSE
    DO I = 1, 5
      IF(Len_Trim(ErrorMessage(I)) < 1) CYCLE
      WRITE(*, *) Trim(ErrorMessage(I))
      WRITE(10, *) Trim(ErrorMessage(I))
    END DO
  END IF
  CLOSE(10)
  DO
    Call CPU_TIME(DelayTime(2))
    IF(DelayTime(2) - DelayTime(1) > 1.0) EXIT
  END DO
  STOP
END SUBROUTINE StopOnError
!
SUBROUTINE Initialize
  USE VARIABLES
  IMPLICIT NONE
  INTEGER :: I, J, K, L
  ALLOCATE (ObsNumAllele(NumLoci), TransRate(NumLoci, 6), LocusError(NumLoci), &
    STAT = I)
  IF(I /= 0) Call StopOnError(0)
  L = Maxval(NumCandidate(1 : 2))
  ALLOCATE(CandidateLogL(L, 2), STAT = I)
  IF(I /= 0) Call StopOnError(0)
  DO L = 1, 2
    MeanNumCandidate(L) = 0
    K = 0
    DO J = 1, SmplSize
      IF(IndivParent(J, L) > 0) CYCLE
      MeanNumCandidate(L) = MeanNumCandidate(L) + &
        NumCandidate(L) - ExcludedParent(J, L)%N
      K = K + 1
    END DO
    IF(K > 0) MeanNumCandidate(L) = MeanNumCandidate(L) / K
  END DO
End SUBROUTINE Initialize
!
SUBROUTINE InitialFSProb
  USE VARIABLES
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L
  EXTERNAL :: CalculatePairProb
!C Generate genotype IDs
  Allocate(GenotypeID2(0 : MaxObsNumallele, 0 : MaxObsNumallele))
  K = MaxObsNumallele
  GenotypeID2(0, 0) = 0
  DO I = 1, MaxObsNumallele
    GenotypeID2(I, I) = I
    GenotypeID2(0, I) = -I
    GenotypeID2(I, 0) = -I
    DO J = I + 1, MaxObsNumallele
      K = K + 1
      GenotypeID2(I, J) = K
      GenotypeID2(J, I) = K
    END DO
  END DO
!C
  DO L = 1, NumLoci
    IF(OtherErrorRate(L) > Small .OR. DropRate(L) > Small)THEN
      LocusError(L) = 1
    ELSE
      LocusError(L) = 0
    END IF
  END DO
  IF(ANY(LocusError(1 : NumLoci) == 0)) THEN
    MarkerErrorFreeQ = .TRUE.
  ELSE
    MarkerErrorFreeQ = .FALSE.
  END IF
!C Convert diploid offspring genotypes to a single number
  Allocate(IndivGType2(NumLoci, SmplSize))
  DO L = 1, NumLoci
    IF(DominantMarker(L) == 1) THEN
      DO I = 1, SmplSize
        IF(IndivGType(1, L, I) < 1)THEN
          IndivGType2(L, I) = 0
        ELSE
          IndivGType2(L, I) = IndivGType(1, L, I)
        END IF
      END DO
    ELSE
      DO I = 1, SmplSize
        IF(IndivGType(1, L, I) < 1)THEN
          IndivGType2(L, I) = 0
        ELSE
          IndivGType2(L, I) = &
            GenotypeID2(IndivGType(1, L, I), IndivGType(2, L, I))
        END IF
      END DO
    END IF
  END DO
!C Initialise variables for known relationships  
  DO I = 1, 2
    IF(NumCandidate(I) < 1) THEN
      NumKnownPrnt(I) = 0
      CYCLE
    END IF
    K = 0
    DO J = 1, SmplSize
      IF(IndivParent(J, I) > 0) K = K + 1
    END DO
    NumKnownPrnt(I) = K
  END DO
  CALL CalculatePairProb
END SUBROUTINE InitialFSProb
!
Subroutine CalculatePairProb
  USE VARIABLES
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, N, IdxError
  REAL(DP) :: X
  INTEGER(I4B), Allocatable :: Prnt(:, :)
  EXTERNAL :: Compute_PODyad_LogL
  DO          
    Allocate(PrbPOPair(SmplSize, 2), STAT = IdxError)
    IF(IdxError /= 0) EXIT
    DO I = 1, 2
      K = NumCandidate(I)
      IF(K < 1) CYCLE
      DO J = 1, SmplSize
        Allocate(PrbPOPair(J, I)%P(K), STAT = IdxError)
        IF(IdxError /= 0) EXIT
      END DO
      IF(IdxError /= 0) EXIT
    END DO
    EXIT
  END DO
!C Initialise variables
  Allocate(Prnt(SmplSize, 2))
  Prnt = IndivParent
  IndivParent = -1
!C Dad offspring likelihood
  write (*, '(/A)') 'Calculating father-offspring dyad likelihood'
  N = 0    
  DO I = 1, NumCandidate(1)
    DO J = 1, SmplSize
      N = N + 1  
      IF(MOD(N, 10000) == 0) WRITE(*, Merge('(A,$)', '(A)  ', MOD(N, 1000000) /= 0)) '.'
      IF(Prnt(J, 1) > 0 .AND. Prnt(J, 1) /= I) THEN
        PrbPOPair(J, 1)%P(I) = Tiny_SP   
        CYCLE
      END IF
      CALL Compute_PODyad_LogL(CndGType(:, :, I, 1), IndivGType(:, :, J), &
        X, DBLE(TINY_SP))
      PrbPOPair(J, 1)%P(I) = REAL(X)
      PrbPOPair(J, 2)%P(I) = PrbPOPair(J, 1)%P(I)
    END DO     
  END DO
  WRITE(*, *) 
!C Copy back the original known parentage
  IndivParent = Prnt
END Subroutine CalculatePairProb
!
SUBROUTINE DataProcess
  USE VARIABLES
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L, M, N, K1, K2, K3
  INTEGER(I4B), ALLOCATABLE :: AlleleList(:), AlleleCount(:)
  REAL(DP) :: X 
  Type(RealVector) :: AlleleFre(NumLoci)
  EXTERNAL :: StopOnError
  Allocate(RawAllele(NumLoci))
!C Check data for known relationships
  DO I = 1, SmplSize
    DO J = 1, 2
      IF(IndivParent(I, J) > 0) THEN
        L = ExcludedParent(I, J)%N
        IF(ANY(IndivParent(I, J) == ExcludedParent(I, J)%List(1 : L))) THEN
          IF(J == 1)THEN
            WRITE(ErrorMessage(1), '(A, I5, A)')'Data error! Offspring ', &
              I, ' has a known father listed as excluded'
          ELSE
            WRITE(ErrorMessage(1), '(A, I5, A)')'Data error! Offspring ', &
              I, ' has a known mother listed as excluded'
          END IF
          Call StopOnError(1)
        END IF
      END IF
    END DO
  END DO
!C Find the observed number of alleles at a locus
  I = (SmplSize + NumCandidate(1) + NumCandidate(2)) * 2
  ALLOCATE(AlleleList(I), AlleleCount(I))
  MaxObsNumAllele = 0
  DO I = 1, NumLoci
    N = 0
    DO J = 1, SmplSize
      IF(ANY(IndivGType(1 : 2, I, J) < 1)) THEN
        IndivGType(1 : 2, I, J) = 0
        CYCLE
      END IF
      DO K = 1, 2
        K1 = 0
        DO K2 = 1, N
          IF(IndivGType(K, I, J) == AlleleList(K2)) THEN
            K1 = K2
            EXIT
          END IF
        END DO
        IF(K1 > 0) THEN
          AlleleCount(K1) = AlleleCount(K1) + 1
        ELSE
          N = N + 1
          AlleleList(N) = IndivGType(K, I, J)
          AlleleCount(N) = 1
        END IF
      END DO
    END DO
    DO M = 1, 2
      K3 = 1
      DO J = 1, NumCandidate(M)
        IF(ANY(CndGType(1 : 2, I, J, M) < 1)) THEN
          CndGType(1 : 2, I, J, M) = 0
          CYCLE
        END IF
        DO K = 1, 2
          K1 = 0
          DO K2 = 1, N
            IF(CndGType(K, I, J, M) == AlleleList(K2)) THEN
              K1 = K2
              EXIT
            END IF
          END DO
          IF(K1 > 0) THEN
            IF(K3 == 1) AlleleCount(K1) = AlleleCount(K1) + 1
          ELSE
            N = N + 1
            AlleleList(N) = CndGType(K, I, J, M)
            IF(K3 == 1) AlleleCount(N) = 1
          END IF
        END DO
      END DO
    END DO
    ObsNumAllele(I) = N
    IF(MaxObsNumAllele < N) MaxObsNumAllele = N
!C Check dominant genotypes
    IF(DominantMarker(I) == 1) THEN
      DO J = 1, SmplSize
        IF(IndivGType(1, I, J) > 2) THEN
          WRITE(ErrorMessage(1),'(A,I6,A,I4,A,I6)') &
            'Individual', J, ' at dominant locus', I, &
            ' has an illegal genotype ', IndivGType(1, I, J)
          WRITE(ErrorMessage(2),'(A)')'Program stops!'
          Call StopOnError(1)
        END IF
      END DO
      DO M = 1, 2
        DO J = 1, NumCandidate(M)
          IF(CndGType(1, I, J, M) > 2) THEN
            IF(M == 1)THEN
              WRITE(ErrorMessage(1),'(A,I6,A,I4,A,I6)') &
                'Candidate father', J, ' at dominant locus', I, &
                ' has an illegal genotype ', CndGType(1, I, J, M)
            ELSE
              WRITE(ErrorMessage(1),'(A,I6,A,I4,A,I6)') &
                'Candidate mother', J, ' at dominant locus', I, &
                ' has an illegal genotype ', CndGType(1, I, J, M)
            END IF
            WRITE(ErrorMessage(2),'(A)')'Program stops!'
            Call StopOnError(1)
          END IF
        END DO
      END DO
    END IF
    IF(DominantMarker(I) == 0) THEN
      J = N
    ELSE
      J = 2
    END IF
    Allocate(RawAllele(I)%R(-1 : J), AlleleFre(I)%R(J))
    DO J = -1, 0
      RawAllele(I)%R(J) = J
    END DO
    IF(DominantMarker(I) == 0) THEN
      X = 1._DP / Sum(AlleleCount(1 : N))
      DO J = 1, ObsNumAllele(I)
        AlleleFre(I)%R(J) = AlleleCount(J) * X
      END DO
      DO J = 1, ObsNumAllele(I)
        RawAllele(I)%R(J) = AlleleList(J)
        WHERE(IndivGType(:, I, :) == AlleleList(J)) IndivGType(:, I, :) = -J
        DO M = 1, 2
          WHERE(CndGType(1 : 2, I, 1 : NumCandidate(M), M) == &
            AlleleList(J)) &
            CndGType(1 : 2, I, 1 : NumCandidate(M), M) = -J
        END DO
      END DO
      IndivGType(:, I, :) = -IndivGType(:, I, :)
      DO M = 1, 2
        CndGType(1 : 2, I, 1 : NumCandidate(M), M) = &
          -CndGType(1 : 2, I, 1 : NumCandidate(M), M)
      END DO
      DO K = 1, SmplSize
        IF(IndivGType(1, I, K) > IndivGType(2, I, K)) &
          IndivGType(1 : 2, I, K) = IndivGType(2 : 1 : -1, I, K)
      END DO
      DO M = 1, 2
        DO K = 1, NumCandidate(M)
          IF(CndGType(1, I, K, M) > CndGType(2, I, K, M)) &
            CndGType(1 : 2, I, K, M) = CndGType(2 : 1 : -1, I, K, M)
        END DO
      END DO
    ELSE               !Dominant markers
      IF(ObsNumAllele(I) == 1) THEN
        IF(AlleleList(1) == 1) THEN
          AlleleCount(2) = 0
        ELSE
          AlleleCount(2) = AlleleCount(1)
          AlleleCount(1) = 0
        END IF
      ELSE
        IF(AlleleList(1) == 2) AlleleCount(1 : 2) = AlleleCount(2 : 1 : -1)
      END IF
      N = Sum(AlleleCount(1 : 2))
      AlleleFre(I)%R(2) = SQRT(DBLE(AlleleCount(2)) / N)
      AlleleFre(I)%R(1) = 1._DP - AlleleFre(I)%R(2)
      DO J = 1, 2
        RawAllele(I)%R(J) = J
      END DO
    END IF
  END DO
! Check whether there is any variation in the genotype data or not
  IF(ALL(ObsNumAllele(1 : NumLoci) < 2))THEN
    WRITE(ErrorMessage(1),'(A)')'Program stops because no variation in geneotypes!'
    WRITE(ErrorMessage(2),'(A)')'All loci have less than 2 observed alleles!'
    Call StopOnError(1)
  END IF
  MaxNumGenotype = MaxObsNumAllele * (MaxObsNumAllele + 1) / 2
  ALLOCATE (ObsAlleleFre(NumLoci, 0 : MaxObsNumAllele))
  DO L = 1, NumLoci
    DO I = 1, ObsNumAllele(L)
      ObsAlleleFre(L, I) = AlleleFre(L)%R(I)
      IF(DominantMarker(L) == 1 .AND. &
          ObsNumAllele(L) == 1 .AND. MaxObsNumAllele > 1)THEN
        ObsAlleleFre(L, 2) = 1._DP - AlleleFre(L)%R(I)
      END IF
    END DO
  END DO
END SUBROUTINE DataProcess
!
SUBROUTINE PhenotypeGivenGenotype
!Determine prob of a phenotype on a genotype at each locus
  USE VARIABLES
  IMPLICIT NONE
  INTEGER :: L
  REAL(DP) :: E(3)
  DO L = 1, NumLoci
    IF(DominantMarker(L) == 0) THEN  !Codominant markers
      IF(ObsNumAllele(L) > 1) THEN
        E(1) = DropRate(L) / (1.0_DP + DropRate(L))
        E(2) = OtherErrorRate(L)
        E(3) = OtherErrorRate(L) / (ObsNumAllele(L) - 1)
        TransRate(L, 1) = (1 - E(2)) ** 2 + E(3) * E(3) - &
                          (E(1) +  E(1)) * (1 - E(2) - E(3)) ** 2
        TransRate(L, 2) = E(3) * (1 - E(2)) + E(1) * (1 - E(2) - E(3)) ** 2
        TransRate(L, 3) = E(3) * (1 - E(2) + E(3))
        TransRate(L, 4) = E(3) * E(3)
        TransRate(L, 5) = (1 - E(2)) ** 2
        TransRate(L, 6) = (E(3) + E(3)) * (1 - E(2))
      ELSE
        TransRate(L, 1 : 6) = (/0._DP, 0._DP, 0._DP, 0._DP, 1._DP, 0._DP/)
      END IF
    ELSE
      IF(ObsNumAllele(L) > 1) THEN
        E(2) = OtherErrorRate(L)
        TransRate(L, 1) = 1._DP - E(2) ** 2
        TransRate(L, 2) = 1._DP - E(2) + E(2) ** 2
        TransRate(L, 3) = E(2) + E(2) - E(2) ** 2
        TransRate(L, 4) = E(2) ** 2
        TransRate(L, 5) = E(2) * (1._DP - E(2))
        TransRate(L, 6) = (1._DP - E(2)) ** 2
      ELSE
        TransRate(L, 1 : 6) = (/1._DP, 0._DP, 0._DP, 0._DP, 0._DP, 0._DP/)
      END IF
    END IF
  END DO
END SUBROUTINE PhenotypeGivenGenotype
!
SUBROUTINE SingleParentLikelihood
  USE VARIABLES
  IMPLICIT NONE
  INTEGER ::  J
  EXTERNAL :: SingleIndivLikelihood
  DO J = 1, NumCandidate(1)
    CALL SingleIndivLikelihood(CndGType(:, :, J, 1), CandidateLogL(J, 1))
    CandidateLogL(J, 2) = CandidateLogL(J, 1)
  END DO
END SUBROUTINE SingleParentLikelihood
!
SUBROUTINE SingleIndivLikelihood(GType, LogL)
  USE VARIABLES
  IMPLICIT NONE
  REAL(DP), INTENT(OUT) :: LogL
  INTEGER(I4B), INTENT(IN) :: GType(2, NumLoci)
  INTEGER(I4B) ::  L, K(2)
  REAL(DP) ::  X
  EXTERNAL :: ProbSingleGType
  LogL = 0._DP
  DO L = 1, NumLoci
    IF(ObsNumAllele(L) < 2) CYCLE
    If(GType(1, L) < 1) CYCLE
    K(1 : 2) = GType(1 : 2, L)
    CALL ProbSingleGType(L, K, X)
    LogL = LogL + Log(X)
  END DO
END SUBROUTINE SingleIndivLikelihood
!
Subroutine ProbSingleGType(L, K, X)
  USE VARIABLES
  IMPLICIT NONE
  INTEGER(I4B), INTENT(IN) :: L, K(2)
  REAL(DP), INTENT(OUT) :: X
  REAL(DP) :: Y(0 : 2)
  IF(DominantMarker(L) /= 1)THEN
    IF(K(1) == K(2))THEN
      X = ObsAlleleFre(L, K(1)) ** 2 * TransRate(L, 5)
      IF(LocusError(L) /= 0)THEN
        X = X + 2._DP * ObsAlleleFre(L, K(1)) * &
           (1._DP - ObsAlleleFre(L, K(1))) * TransRate(L, 2)
        X = X + (1._DP - ObsAlleleFre(L, K(1))) ** 2 * TransRate(L, 4)
      END IF
    ELSE
      Y(1) = ObsAlleleFre(L, K(1))
      Y(2) = ObsAlleleFre(L, K(2))
      X = 2._DP * Y(1) * Y(2) * TransRate(L, 1)
      IF(LocusError(L) /= 0)THEN
        Y(0) = Y(1) + Y(2)
        X = X + (Y(1) * Y(1) + Y(2) * Y(2)) * TransRate(L, 6)
        X = X + 2._DP * Y(0) * (1._DP - Y(0)) * TransRate(L, 3)
        X = X + (1._DP - Y(0)) ** 2 * (TransRate(L, 4) + TransRate(L, 4))
      END IF
    END IF
  ELSE
    IF(K(1) == 1)THEN
      X = ObsAlleleFre(L, 1) ** 2 * TransRate(L, 1) + &
         2._DP * ObsAlleleFre(L, 1) * ObsAlleleFre(L, 2) * TransRate(L, 2)
      IF(LocusError(L) /= 0) X = X + ObsAlleleFre(L, 2) ** 2 * TransRate(L, 3)
    ELSE
      X = ObsAlleleFre(L, 2) ** 2 * TransRate(L, 6)
      IF(LocusError(L) /= 0) X = X + 2._DP * ObsAlleleFre(L, 1) * &
         ObsAlleleFre(L, 2) * TransRate(L, 5) + &
         ObsAlleleFre(L, 1) ** 2 * TransRate(L, 4)
    END IF
  END IF
END Subroutine ProbSingleGType
!
FUNCTION PrbGtoP(Genotype, Phenotype, L)
 ! Calculate prob of a Phenotype given genotype at locus L
  USE VARIABLES
  IMPLICIT NONE
  INTEGER(I4B), INTENT(IN) :: Genotype(2), Phenotype(2), L
  REAL(DP) :: PrbGtoP
  IF (DominantMarker(L) == 0) THEN
    IF(Genotype(1) == Genotype(2))THEN
      IF(ALL(Phenotype(1 : 2) == Genotype(1)))THEN
        PrbGtoP = TransRate(L, 5)
      ELSE IF(ANY(Genotype(1) == Phenotype(1 : 2)))THEN
        PrbGtoP = TransRate(L, 6)
      ELSE
        IF(Phenotype(1) == Phenotype(2))THEN
          PrbGtoP = TransRate(L, 4)
        ELSE
          PrbGtoP = TransRate(L, 4) + TransRate(L, 4)
        END IF
      END IF
    ELSE
      IF(ANY(Genotype(1) == Phenotype(1 : 2)).AND. &
         ANY(Genotype(2) == Phenotype(1 : 2)))THEN
        PrbGtoP = TransRate(L, 1)
      ELSE IF(Phenotype(1) == Phenotype(2) .AND. &
        (Phenotype(2) == Genotype(1) .OR. Phenotype(2) == Genotype(2)))THEN
        PrbGtoP = TransRate(L, 2)
      ELSE IF(ANY(Genotype(1) == Phenotype(1 : 2)) .OR. &
              ANY(Genotype(2) == Phenotype(1 : 2)))THEN
        PrbGtoP = TransRate(L, 3)
      ELSE
        IF(Phenotype(1) == Phenotype(2))THEN
          PrbGtoP = TransRate(L, 4)
        ELSE
          PrbGtoP = TransRate(L, 4) + TransRate(L, 4)
        END IF
      END IF
    END IF
  ELSE
    IF(Phenotype(1) == 1) THEN   !Dominant phenotype
      IF(ALL(Genotype(1 : 2) == 1)) THEN
        PrbGtoP = TransRate(L, 1)
      ELSE IF(ALL(Genotype(1 : 2) /= 1)) THEN
        PrbGtoP = TransRate(L, 3)
      ELSE
        PrbGtoP = TransRate(L, 2)
      END IF
    ELSE
      IF(ALL(Genotype(1 : 2) == 2)) THEN
        PrbGtoP = TransRate(L, 6)
      ELSE IF(ANY(Genotype(1 : 2) == 2)) THEN
        PrbGtoP = TransRate(L, 5)
      ELSE
        PrbGtoP = TransRate(L, 4)
      END IF
    END IF
  END IF
END FUNCTION PrbGtoP
!
SUBROUTINE indexx(n,arr,indx)
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: n
  DOUBLE PRECISION, INTENT(IN) :: arr(n)
  INTEGER, INTENT(OUT) :: indx(n)
  INTEGER M,NSTACK
  DOUBLE PRECISION a
  PARAMETER (M=7,NSTACK=50)
  INTEGER i,indxt,ir,itemp,j,jstack,k,l,istack(NSTACK)
  EXTERNAL :: StackError
  FORALL(j=1 : n) indx(j)=j
  jstack=0
  l=1
  ir=n
1 if(ir-l < M)then
    DO j=l+1,ir
      indxt=indx(j)
      a=arr(indxt)
      DO i=j-1,l,-1
        if(arr(indx(i)) <= a)goto 2
        indx(i+1)=indx(i)
      END DO
      i=l-1
2     indx(i+1)=indxt
    END DO
    if(jstack == 0)return
    ir=istack(jstack)
    l=istack(jstack-1)
    jstack=jstack-2
  else
    k=(l+ir)/2
    itemp=indx(k)
    indx(k)=indx(l+1)
    indx(l+1)=itemp
    if(arr(indx(l)) > arr(indx(ir)))then
      itemp=indx(l)
      indx(l)=indx(ir)
      indx(ir)=itemp
    endif
    if(arr(indx(l+1)) > arr(indx(ir)))then
      itemp=indx(l+1)
      indx(l+1)=indx(ir)
      indx(ir)=itemp
    endif
    if(arr(indx(l)) > arr(indx(l+1)))then
      itemp=indx(l)
      indx(l)=indx(l+1)
      indx(l+1)=itemp
    endif
    i=l+1
    j=ir
    indxt=indx(l+1)
    a=arr(indxt)
3   continue
    i=i+1
    if(arr(indx(i)) < a)goto 3
4   continue
    j=j-1
    if(arr(indx(j)) > a)goto 4
    if(j < i)goto 5
    itemp=indx(i)
    indx(i)=indx(j)
    indx(j)=itemp
    goto 3
5   indx(l+1)=indx(j)
    indx(j)=indxt
    jstack=jstack+2
    if(jstack > NSTACK) CALL StackError(1)
    if(ir-i+1 >= j-l)then
      istack(jstack)=ir
      istack(jstack-1)=i
      ir=j-1
    else
      istack(jstack)=j-1
      istack(jstack-1)=l
      l=i
    endif
  endif
  goto 1
END SUBROUTINE indexx
!
SUBROUTINE indexCh(n, ChLength, indx)
  Use Variables
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: n, ChLength
  INTEGER, INTENT(OUT) :: indx(n)
  Character(Len = ChLength) :: a
  INTEGER, PARAMETER :: M=7,NSTACK=50
  INTEGER :: i,indxt,ir,itemp,j,jstack,k,l,istack(NSTACK)
  FORALL(j=1 : n) indx(j)=j
  jstack=0
  l=1
  ir=n
1 if(ir-l < M)then
    DO j=l+1,ir
      indxt=indx(j)
      a=SortCharacter(indxt)
      DO i=j-1,l,-1
        if(SortCharacter(indx(i)) <= a)goto 2
        indx(i+1)=indx(i)
      END DO
      i=l-1
2     indx(i+1)=indxt
    END DO
    if(jstack == 0)return
    ir=istack(jstack)
    l=istack(jstack-1)
    jstack=jstack-2
  else
    k=(l+ir)/2
    itemp=indx(k)
    indx(k)=indx(l+1)
    indx(l+1)=itemp
    if(SortCharacter(indx(l)) > SortCharacter(indx(ir)))then
      itemp=indx(l)
      indx(l)=indx(ir)
      indx(ir)=itemp
    endif
    if(SortCharacter(indx(l+1)) > SortCharacter(indx(ir)))then
      itemp=indx(l+1)
      indx(l+1)=indx(ir)
      indx(ir)=itemp
    endif
    if(SortCharacter(indx(l)) > SortCharacter(indx(l+1)))then
      itemp=indx(l)
      indx(l)=indx(l+1)
      indx(l+1)=itemp
    endif
    i=l+1
    j=ir
    indxt=indx(l+1)
    a=SortCharacter(indxt)
3   continue
    i=i+1
    if(SortCharacter(indx(i)) < a)goto 3
4   continue
    j=j-1
    if(SortCharacter(indx(j)) > a)goto 4
    if(j < i)goto 5
    itemp=indx(i)
    indx(i)=indx(j)
    indx(j)=itemp
    goto 3
5   indx(l+1)=indx(j)
    indx(j)=indxt
    jstack=jstack+2
    if(jstack > NSTACK) THEN
      WRITE(*, *) 'NSTACK too small in indexx'
      WRITE(*,*)'Program paused. Press ENTER to continue.'
      READ(*,*)
    END IF
    if(ir-i+1 >= j-l)then
      istack(jstack)=ir
      istack(jstack-1)=i
      ir=j-1
    else
      istack(jstack)=j-1
      istack(jstack-1)=l
      l=i
    endif
  endif
  goto 1
END SUBROUTINE indexCh
!
SUBROUTINE StackError(IFlag)
  USE VARIABLES
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: IFlag
  IF(IFlag == 1) THEN
    WRITE(ErrorMessage(1),'(A)') &
      'Program stops due to NSTACK too small in subroutine IndexX'
  END IF
  Call StopOnError(1)
END SUBROUTINE StackError